from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
from itertools import chain
from typing import Dict, Iterable, List, Optional, Type, Union

import networkx as nx

from sbt.package.package import DepConstraints, Package, PackageType


@dataclass
class ThirdPartyPackage:
    name: str
    type: PackageType
    # mapping from source package that use this package to the version the source package depends on
    invert_dependencies: Dict[str, DepConstraints]


class PkgGraph:
    """Representing the dependencies between packages (including third-party packages) in the project.
    The edge between (A, B) represents a dependency relationship that package A uses package B.
    """

    def __init__(self, g: Optional[nx.DiGraph] = None) -> None:
        self.g = g or nx.DiGraph()
        self.pkgs = {}

    @staticmethod
    def from_pkgs(pkgs: Dict[str, Package]) -> PkgGraph:
        """Create a PkgGraph from a dictionary of packages.

        Args:
            pkgs: A dictionary of packages.
        """
        g = nx.DiGraph()
        for pkg in pkgs.values():
            g.add_node(pkg.name, pkg=pkg)

        for pkg in pkgs.values():
            for dep, specs in pkg.dependencies.items():
                if not g.has_node(dep):
                    assert dep not in pkgs
                    g.add_node(
                        dep,
                        pkg=ThirdPartyPackage(dep, pkg.type, {pkg.name: specs}),
                    )
                else:
                    dep_pkg = g.nodes[dep]["pkg"]
                    if isinstance(dep_pkg, ThirdPartyPackage):
                        dep_pkg.invert_dependencies[pkg.name] = specs
                g.add_edge(pkg.name, dep)

        try:
            cycles = nx.find_cycle(g, orientation="original")
            raise ValueError(f"Found cyclic dependencies: {cycles}")
        except nx.NetworkXNoCycle:
            pass

        return PkgGraph(g)

    def iter_pkg(self) -> Iterable[Union[ThirdPartyPackage, Package]]:
        """Iterate over all packages in the graph."""
        return (self.g.nodes[pkg]["pkg"] for pkg in self.g.nodes)

    @lru_cache()
    def dependencies(self, pkg_name: str) -> List[Union[ThirdPartyPackage, Package]]:
        """Return the list of packages that are dependency of the input package.

        Args:
            pkg_name: The package to get the dependencies of.
        """
        nodes = list(nx.dfs_preorder_nodes(self.g, pkg_name))
        assert nodes[0] == pkg_name, "The first node is the package itself"
        nodes = nodes[1:]
        return [self.g.nodes[uid]["pkg"] for uid in nodes]

    def third_party_dependencies(self, pkg_name: str) -> List[ThirdPartyPackage]:
        """Return the list of third-party packages that are dependency of the input package.

        Args:
            pkg_name: The package to get the dependencies of.
        """
        return [
            pkg
            for pkg in self.dependencies(pkg_name)
            if isinstance(pkg, ThirdPartyPackage)
        ]

    def local_dependencies(self, pkg_name: str) -> List[Package]:
        """Return the list of local packages that are dependency of the input package.

        Args:
            pkg_name: The package to get the dependencies of.
        """
        return [
            pkg
            for pkg in self.dependencies(pkg_name)
            if isinstance(pkg, Package)
        ]


import os
from typing import Literal

import click
from loguru import logger

from sbt.config import PBTConfig
from sbt.vcs.git import Git


@click.command()
@click.option(
    "--repo",
    default="",
    help="Specify the multi-repository that we are working with. e.g., https://github.com/binh-vu/pbt",
)
@click.option("--cwd", default=".", help="Override current working directory")
@click.argument("command", type=click.Choice(["sync-dep"]))
def git(repo: str, cwd: str, command: Literal["sync-dep"]):
    cwd = os.path.abspath(cwd)

    cfg = PBTConfig.from_dir(cwd)
    if command == "sync-dep":
        if not cfg.library_path.exists() and len(cfg.dependency_repos) == 0:
            return

        cfg.library_path.mkdir(exist_ok=True, parents=True)

        sync_repos = set(cfg.dependency_repos)
        for subdir in cfg.library_path.iterdir():
            if not Git.is_git_dir(subdir):
                continue
            repo = Git.get_repo(subdir)
            if repo not in sync_repos:
                logger.warning(
                    "Found and skip a git directory in libraries that isn't in the list of dependencies: {}",
                    subdir,
                )
                continue
            sync_repos.remove(repo)
            logger.info("Pull dependency {}", repo)
            Git.pull(subdir, submodules=False)
        for repo in sync_repos:
            logger.info("Clone dependency {}", repo)
            Git.clone(repo, cfg.library_path)
    else:
        raise Exception(f"Invalid command: {command}")



from __future__ import annotations

import os
import shutil
import zipfile
from pathlib import Path
from uuid import uuid4

import click
from loguru import logger

from sbt.commands.install import PASSTHROUGH_ENVS
from sbt.config import PBTConfig
from sbt.misc import exec
from sbt.package.discovery import discover_packages


@click.command()
@click.argument("package", default="")
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "-r",
    "--release",
    is_flag=True,
    help="release mode",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def build(package: str, cwd: str = ".", release: bool = False, verbose: bool = False):
    cwd = os.path.abspath(cwd)
    cfg = PBTConfig.from_dir(cwd)

    # discovery packages
    packages = discover_packages(
        cfg.cwd,
        cfg.cache_dir,
        cfg.ignore_directories,
        cfg.ignore_directory_names,
        ignore_invalid_package=True,
    )

    if package == "":
        # use the package located in the current directory
        for pkg in packages.values():
            if pkg.location == cfg.cwd:
                package = pkg.name
                break
        else:
            raise ValueError(
                f"Cannot find a package in the current directory {cfg.cwd}"
            )

    if package not in packages:
        logger.error(
            f"Cannot find package {package} perhaps the project configuration is incorrect."
        )
        return

    pkg = packages[package]
    outdir = pkg.location / "dist" / str(uuid4())
    outdir.mkdir(parents=True, exist_ok=True)

    cmd = ["maturin", "build"]
    if release:
        cmd.append("-r")
    cmd.extend(["-o", str(outdir)])
    exec(
        cmd,
        cwd=str(pkg.location),
        env=PASSTHROUGH_ENVS,
    )

    (whl_file,) = [x for x in outdir.glob("*.whl")]
    with zipfile.ZipFile(whl_file, "r") as zip_ref:
        zip_ref.extractall(outdir)

    pkg_name = pkg.name.replace("-", "_")
    pkg_dir = outdir / pkg_name
    if not pkg_dir.exists():
        for name in pkg.include:
            pkg_name = name.replace("-", "_")
            pkg_dir = outdir / pkg_name
            if pkg_dir.exists():
                break
        else:
            raise RuntimeError(f"Cannot find the package directory in {outdir}")

    for file in pkg_dir.iterdir():
        if file.name.endswith(".so") or file.name.endswith(".dylib"):
            dest_file = pkg.location / pkg_name / file.name
            if dest_file.exists():
                dest_file.unlink()
            shutil.move(file, pkg.location / pkg_name)

    shutil.rmtree(outdir)



import importlib.metadata
import sys

import click
from loguru import logger

from sbt.commands.build import build
from sbt.commands.git import git
from sbt.commands.install import add, install
from sbt.registry.pypi import PyPI

try:
    version = importlib.metadata.version("pbt-simple")
except importlib.metadata.PackageNotFoundError:
    version = "0.0.0"


def check_upgrade():
    latest_version = PyPI.get_instance().get_latest_version("pbt-simple")
    if latest_version is not None and version != latest_version:
        logger.warning(
            f"You are using an outdated version of sbt (pbt-simple). The latest version is {latest_version}, while you are using {version}."
        )
    else:
        logger.trace("You are using the latest version of sbt (pbt-simple).")


@click.group(
    help=f"SBT ({version}) -- a simple python build tool for multi-projects that supports poetry and maturin (thus support extension module written in PYO3 (Rust))"
)
@click.version_option(version)
def cli():
    check_upgrade()


cli.add_command(install)
cli.add_command(add)
cli.add_command(build)
cli.add_command(git)

if __name__ == "__main__":
    cli()



from __future__ import annotations

import os
import re
from operator import attrgetter
from pathlib import Path
from typing import cast

import semver
from loguru import logger
from tomlkit.api import loads

from sbt.misc import InvalidPackageError
from sbt.package.package import (
    DepConstraint,
    DepConstraints,
    Package,
    PackageType,
    VersionSpec,
)


def discover_packages(
    root: Path,
    cache_dir: Path,
    ignore_dirs: set[Path],
    ignore_dirnames: set[str],
    ignore_invalid_package: bool = False,
):
    """Find all packages in the given directory and manually linked/installed packages."""
    candidate_pyprojects = get_candidate_pyprojects(root, ignore_dirs, ignore_dirnames)

    # mapping from package directory to its configuration
    pkgs: dict[str, Package] = {}

    for loc in candidate_pyprojects:
        try:
            pkg = parse_pep518_pyproject(loc)
        except InvalidPackageError as e:
            if not ignore_invalid_package:
                raise e
            logger.warning(f"An package at {loc} is invalid. Ignore it. Error: {e}")
            continue

        if pkg.name in pkgs:
            raise RuntimeError(
                f"Duplicate package {pkg.name}: found in {pkg.location} and {pkgs[pkg.name].location}"
            )
        pkgs[pkg.name] = pkg

    # now find the manually linked/installed packages
    for pkg in list(pkgs.values()):
        pkg_cache_dir = cache_dir / pkg.name
        pkg_cache_dir.mkdir(exist_ok=True, parents=True)

        for loc in pkg.find_manually_installed_dependencies(pkg_cache_dir):
            try:
                pkg = parse_pep518_pyproject(loc)
            except InvalidPackageError as e:
                if not ignore_invalid_package:
                    raise e
                logger.warning(f"An package at {loc} is invalid. Ignore it. Error: {e}")
                continue

            if pkg.name in pkgs:
                raise RuntimeError(
                    f"Duplicate package {pkg.name}: found in {pkg.location} and {pkgs[pkg.name].location}"
                )
            pkgs[pkg.name] = pkg

    return pkgs


def get_candidate_pyprojects(
    root: Path, ignore_dirs: set[Path], ignore_dirnames: set[str]
):
    outs = []
    root = root.resolve()

    stack = [root]
    while len(stack) > 0:
        dir = stack.pop()
        if (
            dir.name.startswith(".")
            or dir.name in ignore_dirnames
            or dir in ignore_dirs
        ):
            continue
        if (dir / "pyproject.toml").exists():
            outs.append(dir)
        stack.extend([subdir for subdir in dir.iterdir() if subdir.is_dir()])

    return outs


def parse_pep518_pyproject(loc: Path) -> Package:
    """Parse project metadata from the directory.

    Arguments:
        pyproject_file: path to the pyproject.toml file

    Returns: dict, force the type to dictionary to silence the annoying type checker due to tomlkit
    """
    try:
        with open(loc / "pyproject.toml", "r") as f:
            cfg = cast(dict, loads(f.read()))
    except:
        logger.error("Invalid TOML file: {}", loc / "pyproject.toml")
        raise

    backend = cfg["build-system"]["build-backend"]
    if backend == "maturin":
        return parse_maturin_project(cfg, loc)
    if backend == "poetry.core.masonry.api":
        return parse_poetry_project(cfg, loc)
    raise NotImplementedError(f"Unsupported build backend: {backend}")


def parse_maturin_project(cfg: dict, loc: Path) -> Package:
    name = cfg["project"]["name"]
    version = cfg["project"]["version"]

    dependencies = {}
    for item in cfg["project"]["dependencies"]:
        k, specs = parse_pep518_dep_spec(item)
        dependencies[k] = specs

    for extra, extra_deps in cfg["project"]["optional-dependencies"].items():
        for item in extra_deps:
            if item.find(" ") == -1:
                # it should be a self reference dependency, so we just ignore it
                k, extras = parse_pep518_pkgname_with_extra(item)
                assert k == name, f"{k} doesn't have version (because of no space -- if it does have version, add space to format it nicely), so it must be self-reference"
                continue

            k, specs = parse_pep518_dep_spec(item)
            # the dependency sometimes is self reference, so we just ignore it
            if k == name:
                continue
            dependencies[k] = specs

    # not supported yet in pep-621
    # https://peps.python.org/pep-0621/#specify-files-to-include-when-building
    include = []
    if "tool" in cfg and "maturin" in cfg["tool"]:
        if "include" in cfg["tool"]["maturin"]:
            include = cfg["tool"]["maturin"]["include"]

    return Package(name, version, loc, PackageType.Maturin, include, dependencies)


def parse_poetry_project(cfg: dict, loc: Path) -> Package:
    name = cfg["tool"]["poetry"]["name"]
    version = cfg["tool"]["poetry"]["version"]

    dependencies = {}

    for cfgkey in ["dependencies", "dev-dependencies"]:
        for k, vs in cfg["tool"]["poetry"].get(cfgkey, {}).items():
            if not isinstance(vs, list):
                vs = [vs]
            dependencies[k] = sorted(
                (parse_poetry_dep_spec(v) for v in vs),
                key=attrgetter("constraint"),
            )

    # see https://python-poetry.org/docs/pyproject/#include-and-exclude
    # and https://python-poetry.org/docs/pyproject/#packages
    include = cfg["tool"]["poetry"].get("include", [])
    for pkg_cfg in cfg["tool"]["poetry"].get("packages", []):
        include.append(os.path.join(pkg_cfg.get("from", ""), pkg_cfg["include"]))
    include = sorted(set(include))

    return Package(name, version, loc, PackageType.Poetry, include, dependencies)


def parse_pep518_pkgname_with_extra(name: str) -> tuple[str, list[str]]:
    """Parse a spec containing extra dependencies: `<name>([(<extra>,)+])?"""
    m = re.match(r"([^\[]+)(?:\[(.+)\])?$", name)
    assert m is not None, name

    # TODO: this will throw an error
    name = m.group(1)
    if m.group(2) is not None:
        extras = m.group(2).split(",")
        assert len(extras) > 0
        return name, extras
    return name, []


def parse_pep518_dep_spec(
    spec: str, allow_self_reference: bool = False
) -> tuple[str, DepConstraints]:
    """Parse a dependency specification.

    It does not support PEP508 but only a simple syntax: `<name>([(<extra>,)+])? <version_rule>`.
    Note: the space is important.
    """
    name, version = spec.split(" ", 1)
    name, extras = parse_pep518_pkgname_with_extra(name)
    if len(extras) > 0:
        origin_spec = {"extras": extras}
    else:
        origin_spec = None

    # do it here to make sure we can parse this version
    parse_version_spec(version)
    constraint = f"python=* markers="
    return name, [
        DepConstraint(
            version_spec=version, constraint=constraint, origin_spec=origin_spec
        )
    ]


def parse_poetry_dep_spec(spec: str | dict) -> DepConstraint:
    if isinstance(spec, str):
        constraint = f"python=* markers="
        return DepConstraint(version_spec=spec, constraint=constraint)
    elif isinstance(spec, dict):
        if "version" not in spec:
            if "url" in spec:
                # try to figure out the version from the URL if possible, otherwise, use 1.0.0
                m = re.search(r"\d+.\d+.\d+", spec["url"])
                if m is not None:
                    version_spec = f"=={m.group()}"
                else:
                    version_spec = "==1.0.0"
            else:
                raise NotImplementedError(
                    f"Not support specify dependency outside of Pypi yet. But found spec {spec}"
                )
        else:
            version_spec = spec["version"]

        constraint = (
            f"python={spec.get('python', '*')} markers={spec.get('markers', '')}"
        )
        origin_spec = spec.copy()
        if "version" in origin_spec:
            origin_spec.pop("version")

        return DepConstraint(
            version_spec=version_spec,
            constraint=constraint,
            version_spec_field="version",
            origin_spec=origin_spec,
        )


def parse_version_spec(
    rule: str,
) -> VersionSpec:
    """Parse the given version rule to get lowerbound and upperbound (exclusive)

    Example:
        - "^1.0.0" -> (1.0.0, 2.0.0)
        - ">= 1.0.0" -> (1.0.0, None)
        - ">= 1.0.0, < 2.1.3" -> (1.0.0, 2.1.3)
    """
    m = re.match(
        r"(?P<op1>\^|~|>|>=|==|<|<=)? *(?P<version1>[^ ,\^\~>=<]+)(?:(?:(?: *, *)|(?: +))(?P<op2>\^|~|>|>=|==|<|<=) *(?P<version2>[^ ,\^\~>=<]+))?",
        rule.strip(),
    )
    assert (
        m is not None
    ), f"The constraint is too complicated to handle for now: `{rule}`"

    op1, version1 = m.group("op1"), m.group("version1")
    op2, version2 = m.group("op2"), m.group("version2")

    if op1 == "":
        op1 = "=="

    lowerbound = parse_version(version1)
    if op1 == "^":
        assert version2 is None
        # special case for 0 following the nodejs way (I can't believe why)
        # see more: https://nodesource.com/blog/semver-tilde-and-caret/
        if lowerbound.major == 0:
            if lowerbound.minor == 0:
                upperbound = lowerbound.bump_patch()
            else:
                upperbound = lowerbound.bump_minor()
        else:
            upperbound = lowerbound.bump_major()
        spec = VersionSpec(
            lowerbound=lowerbound,
            upperbound=upperbound,
            is_lowerbound_inclusive=True,
            is_upperbound_inclusive=False,
        )
    elif op1 == "~":
        assert version2 is None
        if m.group("version1").isdigit():
            # only contains major version
            upperbound = lowerbound.bump_major()
        else:
            upperbound = lowerbound.bump_minor()
        spec = VersionSpec(
            lowerbound=lowerbound,
            upperbound=upperbound,
            is_lowerbound_inclusive=True,
            is_upperbound_inclusive=False,
        )
    elif op1 == "==":
        assert version2 is None
        upperbound = lowerbound
        spec = VersionSpec(
            lowerbound=lowerbound,
            upperbound=upperbound,
            is_lowerbound_inclusive=True,
            is_upperbound_inclusive=True,
        )
    else:
        upperbound = parse_version(version2) if version2 is not None else None
        if op1 == "<" or op1 == "<=":
            op1, op2 = op2, op1
            lowerbound, upperbound = upperbound, lowerbound
        spec = VersionSpec(
            lowerbound=lowerbound,
            upperbound=upperbound,
            is_lowerbound_inclusive=op1 == ">=",
            is_upperbound_inclusive=op2 == "<=",
        )
    return spec


def parse_version(version: str) -> semver.VersionInfo:
    m = re.match(
        r"^(?P<major>\d+)(?P<minor>\.\d+)?(?P<patch>\.\d+)?(?P<rest>[^\d].*)?$",
        version,
    )
    assert (
        m is not None
    ), f"Current parser is not able to parse version: `{version}` yet"

    if m is not None:
        parts = [
            m.group("major"),
            m.group("minor") or ".0",
            m.group("patch") or ".0",
            m.group("rest") or "",
        ]
        if not parts[-1].startswith("-") and parts[-1] != "":
            # add hyphen to make it compatible with semver package.
            # e.g. 21.11b1 -> 21.11.0-b1
            parts[-1] = "-" + parts[-1]
        version = "".join(parts)

    return semver.VersionInfo.parse(version)



from __future__ import annotations

import os
import shutil
from collections import defaultdict
from pathlib import Path
from typing import Any, Optional, cast

import click
from loguru import logger
from tomlkit.api import document, dumps, inline_table, loads, nl, table
from tomlkit.items import Array, KeyType, SingleKey, Trivia
from tqdm.auto import tqdm

from sbt.config import PBTConfig
from sbt.misc import (
    ExecProcessError,
    IncompatibleDependencyError,
    NewEnvVar,
    exec,
    mask_file,
    venv_path,
)
from sbt.package.discovery import (
    discover_packages,
    parse_pep518_pyproject,
    parse_version,
    parse_version_spec,
)
from sbt.package.graph import PkgGraph
from sbt.package.package import DepConstraint, DepConstraints, Package

# environment variables that will be passed to the subprocess
PASSTHROUGH_ENVS = [
    "PATH",
    "CC",
    "CXX",
    "LIBCLANG_PATH",
    "LD_LIBRARY_PATH",
    "C_INCLUDE_PATH",
    "CPLUS_INCLUDE_PATH",
    "CARGO_HOME",
    "RUSTUP_HOME",
]


@click.command()
@click.argument("package", default="")
@click.option("--cwd", default=".", help="Override current working directory")
@click.option(
    "--ignore-invalid-pkg",
    is_flag=True,
    help="whether to ignore invalid packages",
)
@click.option(
    "--ignore-invalid-dependency",
    is_flag=True,
    help="whether to ignore invalid dependencies",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def install(
    package: str,
    cwd: str = ".",
    ignore_invalid_pkg: bool = False,
    ignore_invalid_dependency: bool = False,
    verbose: bool = False,
):
    """Install a package and its local dependencies in editable mode"""
    cwd = os.path.abspath(cwd)
    cfg = PBTConfig.from_dir(cwd)

    # discovery packages
    packages = discover_packages(
        cfg.cwd,
        cfg.cache_dir,
        cfg.ignore_directories,
        cfg.ignore_directory_names,
        ignore_invalid_package=ignore_invalid_pkg,
    )
    loc2pkg = {pkg.location: pkg for pkg in packages.values()}
    if package == "":
        # use the package located in the current directory
        if cfg.cwd in loc2pkg:
            package = loc2pkg[cfg.cwd].name
        else:
            raise ValueError(
                f"Cannot find a package in the current directory {cfg.cwd}"
            )

    install_pkg(packages[package], packages, cfg, ignore_invalid_dependency, None)


@click.command()
@click.argument("dependency")
@click.option("--package", default="", help="The target package to add dependency to")
@click.option("--cwd", default=".", help="Override current working directory")
@click.option("--no-dep-dep", is_flag=True, help="Do not install dependencies of the dependency")
@click.option(
    "--ignore-invalid-pkg",
    is_flag=True,
    help="whether to ignore invalid packages",
)
@click.option(
    "--ignore-invalid-dependency",
    is_flag=True,
    help="whether to ignore invalid dependencies",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def add(
    dependency: str,
    package: str = "",
    cwd: str = ".",
    no_dep_dep: bool = False,
    ignore_invalid_pkg: bool = False,
    ignore_invalid_dependency: bool = False,
    verbose: bool = False,
):
    """Add a local package as a dependency to the target package"""
    cwd = os.path.abspath(cwd)
    cfg = PBTConfig.from_dir(cwd)

    # discovery packages
    packages = discover_packages(
        cfg.cwd,
        cfg.cache_dir,
        cfg.ignore_directories,
        cfg.ignore_directory_names,
        ignore_invalid_package=ignore_invalid_pkg,
    )
    loc2pkg = {pkg.location: pkg for pkg in packages.values()}
    if package == "":
        # use the package located in the current directory
        if cfg.cwd in loc2pkg:
            package = loc2pkg[cfg.cwd].name
        else:
            raise ValueError(
                f"Cannot find a package in the current directory {cfg.cwd}"
            )

    # step 1: get the dependency package and add its to the list of discover packages
    if dependency not in packages:
        # dependency is a path
        deppkg = parse_pep518_pyproject(Path(dependency))
        packages[deppkg.name] = deppkg
    else:
        deppkg = packages[dependency]

    # step 2: update the list of manually installed
    pkg_cache_dir = cfg.pkg_cache_dir(packages[package])
    manual_deps = packages[package].find_manually_installed_dependencies(pkg_cache_dir)
    manual_deps.append(deppkg.location)
    packages[package].save_manually_installed_dependencies(
        pkg_cache_dir, sorted(set(manual_deps))
    )

    # step 3: gather dependencies of the package -- include the one from manually installed packages
    if no_dep_dep:
        pkg_venv_path = venv_path(
            packages[package].name,
            packages[package].location,
            cfg.python_virtualenvs_path,
            cfg.get_python_path(),
        )
        install_bare_pkg(packages[package], cfg, pkg_venv_path)
    else:
        install_pkg(packages[package], packages, cfg, ignore_invalid_dependency, [deppkg])


def install_pkg(
    target_pkg: Package,
    packages: dict[str, Package],
    cfg: PBTConfig,
    ignore_invalid_dependency: bool,
    local_dep_pkgs: Optional[list[Package]],
):
    loc2pkg = {pkg.location: pkg for pkg in packages.values()}

    # step 0: gather dependencies of the package -- include the one from manually installed packages
    target_pkg.dependencies.update(
        {
            (p := loc2pkg[loc]).name: [
                DepConstraint(version_spec="=={}".format(p.version))
            ]
            for loc in target_pkg.find_manually_installed_dependencies(
                cfg.pkg_cache_dir(target_pkg)
            )
        }
    )
    pkg_graph = PkgGraph.from_pkgs(packages)

    thirdparty_pkgs: dict[str, tuple[set[str], DepConstraints]] = {}
    invalid_thirdparty_pkgs: set[str] = set()

    for thirdparty_pkg in pkg_graph.third_party_dependencies(target_pkg.name):
        assert thirdparty_pkg.name not in thirdparty_pkgs
        depspecs = None
        for p, pc in thirdparty_pkg.invert_dependencies.items():
            if depspecs is None:
                depspecs = pc
            else:
                try:
                    depspecs = find_common_specs(depspecs, pc)
                except IncompatibleDependencyError:
                    logger.error(
                        "Encounter an incompatible dependency {}. Found it in:\n{}",
                        thirdparty_pkg.name,
                        "\n".join(
                            f"\t- {packages[pkgname].location}"
                            for pkgname in thirdparty_pkg.invert_dependencies.keys()
                        ),
                    )

                    if ignore_invalid_dependency:
                        invalid_thirdparty_pkgs.add(thirdparty_pkg.name)
                    else:
                        raise
        assert depspecs is not None
        thirdparty_pkgs[thirdparty_pkg.name] = (
            set(thirdparty_pkg.invert_dependencies.keys()),
            depspecs,
        )

    # now install the target package
    # step 1: gather all dependencies in one file and install it.
    install_pkg_dependencies(
        target_pkg,
        {depname: depspecs for depname, (_, depspecs) in thirdparty_pkgs.items()},
        cfg,
    )

    # step 2: install all local packages in editable mode
    if local_dep_pkgs is None:
        local_dep_pkgs = [pkg for pkg in pkg_graph.local_dependencies(target_pkg.name)]

    pkg_venv_path = venv_path(
        target_pkg.name,
        target_pkg.location,
        cfg.python_virtualenvs_path,
        cfg.get_python_path(),
    )
    
    if len(local_dep_pkgs) > 0:
        logger.info(
            "Installing local packages: {}", ", ".join(pkg.name for pkg in local_dep_pkgs)
        )
        for pkg in tqdm(local_dep_pkgs):
            install_bare_pkg(pkg, cfg, pkg_venv_path)

    # step 3: check if we need to build and install any extension module
    # TODO: implement this -- for now, users can use the `build` command to build manually
    logger.info("Finished installing package {} to env {}", target_pkg.name, pkg_venv_path)


def install_bare_pkg(pkg: Package, cfg: PBTConfig, virtualenv: Optional[Path] = None):
    """Install a package without any dependencies in editable mode"""
    with mask_file(pkg.location / "pyproject.toml"), mask_file(
        pkg.location / "poetry.lock"
    ):
        with open(pkg.location / "pyproject.toml", "w") as f:
            doc = document()

            tbl = table()
            tbl.add("name", pkg.name)
            tbl.add("version", pkg.version)
            tbl.add("description", "")
            tbl.add("authors", [])
            if sum(int(x != pkg.name) for x in pkg.include) > 0:
                tbl.add("packages", [{"include": x} for x in pkg.include])

            doc.add(SingleKey("tool.poetry", t=KeyType.Bare), tbl)

            tbl = table()
            tbl.add("requires", ["poetry-core>=1.0.0"])
            tbl.add("build-backend", "poetry.core.masonry.api")
            doc.add(nl())
            doc.add("build-system", tbl)

            f.write(dumps(doc))

        install_poetry_package(pkg, cfg, virtualenv, quiet=True)


def install_pkg_dependencies(
    pkg: Package,
    deps: dict[str, DepConstraints],
    cfg: PBTConfig,
    virtualenv: Optional[Path] = None,
):
    with open(cfg.pkg_cache_dir(pkg) / "pyproject.modified.toml", "w") as f:
        doc = document()

        tbl = table()
        tbl.add("name", pkg.name)
        tbl.add("version", pkg.version)
        tbl.add("description", "")
        tbl.add("authors", [])
        
        if sum(int(x != pkg.name) for x in pkg.include) > 0:
            tbl.add("packages", [{"include": x} for x in pkg.include])

        doc.add(SingleKey("tool.poetry", t=KeyType.Bare), tbl)

        tbl = table()
        tbl.add("python", f"=={cfg.get_python_version()}")
        for dep, specs in deps.items():
            tbl.add(dep, serialize_dep_specs(specs))
        doc.add(nl())
        doc.add(SingleKey("tool.poetry.dependencies", t=KeyType.Bare), tbl)

        tbl = table()
        tbl.add("requires", ["poetry-core>=1.0.0"])
        tbl.add("build-backend", "poetry.core.masonry.api")
        doc.add(nl())
        doc.add("build-system", tbl)

        f.write(dumps(doc))

    try:
        os.rename(
            pkg.location / "pyproject.toml",
            cfg.pkg_cache_dir(pkg) / "pyproject.origin.toml",
        )
        if (pkg.location / "poetry.lock").exists():
            os.rename(
                pkg.location / "poetry.lock",
                cfg.pkg_cache_dir(pkg) / "poetry.origin.lock",
            )
        shutil.copy(
            cfg.pkg_cache_dir(pkg) / "pyproject.modified.toml",
            pkg.location / "pyproject.toml",
        )
        if (cfg.pkg_cache_dir(pkg) / "poetry.modified.lock").exists():
            shutil.copy(
                cfg.pkg_cache_dir(pkg) / "poetry.modified.lock",
                pkg.location / "poetry.lock",
            )

        install_poetry_package(pkg, cfg, virtualenv)
    finally:
        os.rename(
            cfg.pkg_cache_dir(pkg) / "pyproject.origin.toml",
            pkg.location / "pyproject.toml",
        )
        if (pkg.location / "poetry.lock").exists():
            os.rename(
                pkg.location / "poetry.lock",
                cfg.pkg_cache_dir(pkg) / "poetry.modified.lock",
            )
        if (cfg.pkg_cache_dir(pkg) / "poetry.origin.lock").exists():
            os.rename(
                cfg.pkg_cache_dir(pkg) / "poetry.origin.lock",
                pkg.location / "poetry.lock",
            )


def install_poetry_package(
    pkg: Package,
    cfg: PBTConfig,
    virtualenv: Optional[Path] = None,
    quiet: bool = False,
):
    if virtualenv is None:
        virtualenv = venv_path(
            pkg.name,
            pkg.location,
            cfg.python_virtualenvs_path,
            cfg.get_python_path(),
        )

    env: list[str | NewEnvVar] = [x for x in PASSTHROUGH_ENVS if x != "PATH"]
    for k, v in get_virtualenv_environment_variables(virtualenv).items():
        env.append({"name": k, "value": v})

    if (pkg.location / "poetry.lock").exists():
        try:
            exec(
                "poetry check --lock" + (" -q" if quiet else ""),
                cwd=pkg.location,
                env=env,
            )
        except ExecProcessError:
            logger.debug(
                "Package {} poetry.lock is inconsistent with pyproject.toml, updating lock file...",
                pkg.name,
            )
            exec(
                "poetry lock --no-update" + (" -q" if quiet else ""),
                cwd=pkg.location,
                capture_stdout=False,
                env=env,
            )

    exec(
        f"poetry install" + (" -q" if quiet else ""),
        cwd=pkg.location,
        capture_stdout=False,
        env=env,
    )


def get_virtualenv_environment_variables(virtualenv: Path) -> dict:
    return {
        "VIRTUAL_ENV": str(virtualenv),
        "PATH": str(virtualenv / "bin") + os.pathsep + os.environ.get("PATH", ""),
    }


def find_common_specs(
    depspecs: DepConstraints, another_depspecs: DepConstraints
) -> DepConstraints:
    """Find the common specs between two dependencies."""
    specs = {x.constraint or "": x for x in depspecs}
    anotherspecs = {x.constraint or "": x for x in another_depspecs}

    if not (len(specs) == len(depspecs) and len(anotherspecs) == len(another_depspecs)):
        raise IncompatibleDependencyError(
            f"Two dependencies have duplicated specs: {depspecs} and {another_depspecs}"
        )

    if specs.keys() != anotherspecs.keys():
        raise IncompatibleDependencyError(
            f"Two dependencies have different number of specs: {depspecs} and {another_depspecs}"
        )

    newspecs = []
    for constraint, spec in specs.items():
        anotherspec = anotherspecs[constraint]

        specver = parse_version_spec(spec.version_spec)
        anotherspecver = parse_version_spec(anotherspec.version_spec)

        # we should not have ValueError exception because we have checked the compatibility
        try:
            specver = specver.intersect(anotherspecver)
        except ValueError:
            raise IncompatibleDependencyError(
                f"Two dependencies have incompatible specs: {depspecs} and {another_depspecs}"
            )

        newspecs.append(
            DepConstraint(
                specver.to_pep508_string(),
                constraint,
                spec.version_spec_field,
                spec.origin_spec,
            )
        )
    return newspecs


def serialize_dep_specs(specs: DepConstraints) -> Array:
    items = []
    for spec in specs:
        if spec.origin_spec is None:
            item = spec.version_spec
        else:
            item = inline_table()
            item[cast(str, spec.version_spec_field)] = spec.version_spec
            for k, v in spec.origin_spec.items():
                item[k] = v
        items.append(item)

    if len(items) == 1:
        return items[0]
    return Array(items, Trivia(), multiline=True)


# def load_external_dependencies(pkg: Package, cfg: PBTConfig, packages: dict[str, Package]) -> dict[str, Package]:
